# Copyright(c) 1997-2001 Inger Jorgensen
# Copyright(c) 2000-2001 Association of Universities for Research in Astronomy, Inc.

procedure gemseeing(image)

# Semi-automatic method for determination of image quality for 
# imaging data. 
# Output: direct fwhm, moffat fwhm, eed50, eed85, strehl ratio
#
# Suggested parameter settings:
#  Seeing  0.4-0.8 arcsec: fwhmin=0.6, radius=0, fwhmmin=0, fwhmmax=0
#  Seeing    2-4   arcsec: fwhmin=3, radius=0, fwhmmin=0, fwhmmax=0
# Default values in database gemtools$gemseeing.dat reflects expected
# seeing with various instruments.
# 
# Known bugs/limitations:
#  strehl ratio output is not reliable in crowded fields
#  a MEF file with valid extensions may have NEXTEND undefinded in PHU
#  this causes gemseeing to crash - and is not a valid FITS header
#  though it works for most tasks.
#
# Version  Sept 1, 2000  IJ
#          Jan 22, 2001  IJ  changed all sat to 1 object instead of <3

char  image     {prompt="Image"}
char  coords    {"default",prompt="daofind output file"}
char  psffile   {"default",prompt="File for psfmeasure output"}
bool  fl_inter  {no,prompt="Interactive use"}
bool  fl_useall {no,prompt="Use all objects if N>maxnobj"}
int   maxnobj   {100,min=5,prompt="Approximate maximum number of objects"}
bool  fl_update {yes,prompt="Update image keywords"}
bool  fl_keep   {yes,prompt="Keep output tables"}
bool  fl_strehl {yes,prompt="Derive Strehl ratio if defined for instrument"}
bool  fl_overwrite {no,prompt="Overwrite results from previous use"}
bool  fl_cleed  {yes,prompt="Clean measurements of EED of outliers"}
char  key_inst  {"INSTRUME",prompt="Keyword for instrument"}
char  key_camera {"CAMERA",prompt="Keyword for NIRI camera"}
char  key_ron   {"RDNOISE",prompt="Keyword for read out noise in electrons"}
char  key_gain  {"GAIN",prompt="Keyword for gain in electrons/ADU"}
char  key_filter {"FILTER",prompt="Keyword for filter"}
char  instrument {"",prompt="Instrument"}
char  camera    {"",prompt="NIRI camera"}
real  ron       {1.,min=0.,prompt="Read out noise in electrons"}
real  gain      {1.,min=0.00001,prompt="Gain in e-/ADU"}
char  filter    {"",prompt="Filter name"}
bool  fl_database {yes,prompt="Use database values for parameters"}
char  database  {"gemtools$gemseeing.dat",prompt="Database to use"}
char  stdatabase {"gemtools$gemseeing_strehl.dat",prompt="Database for Strehl ratio"}
real  pixscale  {0.5,min=0.00001,prompt="Pixel scale arcsec/pixel [database]"}
real  fwhmin    {1.,min=0.001,prompt="Input FWHM PSF in arcsec (daofind) [database]"}
real  datamax   {40000.,prompt="Maximum valid data value [database]"}
real  radius    {0.,min=0.,prompt="Radius in pixels for fit [database]"}
real  buffer    {8.,min=0.,prompt="Buffer zone in pixels for sky [database]"}
real  width     {2.,min=1.,prompt="Width in pixels for sky [database]"}
real  fwhmnonao {0.5,min=0.05,prompt="App non-AO FWHM in arcsec if using AO"}
real  sigthres  {10.,min=0.,prompt="Threshold in sigma of sky (daofind)"}
real  threshold {INDEF,min=0.,prompt="Threshold for daofind (daofind)"}
int   niter     {3,min=1,prompt="Number of iterations (psfmeasure)"}
real  mdaofaint   {-2.0,max=0.,prompt="Faintest magnitude from daofind"}
real  fwhmmin   {0.,min=0.,prompt="Minimim fwhm in arcsec"}
real  fwhmmax   {0.,min=0.,prompt="Maximum fwhm in arcsec"}
real  sharplow  {0.3,prompt="Minimum sharp from daofind"}
real  sharphigh {1.0,prompt="Maximum sharp from daofind"}
real  sharpsig  {3.,min=0.,prompt="Sigma selection on sharp from daofind"}
real  dmagsig   {2.,min=0.,prompt="Sigma selection on magnitude difference"}
real  fwhmsig   {2.5,min=0.,prompt="Sigma selection on FWHM"}
char  logfile   {"gemseeing.log",prompt="Logfile"}
bool  verbose   {yes,prompt="Print intermediate results"}
int   status    {0,prompt="Exit status (0=good)"}

begin

char l_image, l_coords, l_psffile, l_database, l_stdatabase
char l_key_ron, l_key_gain, l_key_inst, l_inst, l_camera, l_key_camera
char l_key_filter, l_filter
char l_logfile
real l_pixscale, l_fwhmin, l_ron, l_gain, l_threshold, l_sigthres
real l_fwhmmin, l_fwhmmax, l_radius, l_buffer, l_width, l_fwhmsig
real l_mdaofaint, l_datamax, l_sharpsig, l_dmagsig, l_sharplow, l_sharphigh
real l_fwhmnonao
bool l_fl_inter, l_fl_overwrite, l_fl_cleed
bool l_fl_update, l_fl_useall, l_fl_keep, l_fl_database, l_fl_strehl

int  l_niter, l_maxnobj, ii
real out_npsf, out_fwhm, out_mfwhm, out_eed50, out_eed85, out_eps, out_pos
real out_strehl
real rms_npsf, rms_fwhm, rms_mfwhm, rms_eed50, rms_eed85, rms_eps, rms_pos
real rms_strehl, x_psf, y_psf

char tmpcoo, tmpdao, tmpime, tmpcd, tmpdat, tmpfit, tmpmom, tmpgra, junk
char l_expression, s_image, l_section, junk1, junk2, t_image, l_sciext, l_save
real l_level, l_skysigma, l_npeak, l_wavelength
int  l_Xmax, l_Ymax, l_obj, n_i, l_ext, l_col, l_nsec
bool l_verbose, l_fl_ext, l_fl_disp, l_fl_one
struct l_struct

l_verbose=verbose
l_image=image ; l_coords=coords ; l_psffile=psffile
# Make image name without section if necessary, s_image is the name
# without the section.  Cannot check access with one statement
# if image section is present.
# l_section contains MEF extension if present and section
s_image=l_image ; t_image=l_image ; l_section=""
if( stridx("[",l_image) != 0 ) { 
   s_image=substr(l_image,1,(stridx("[",l_image)-1) )
   l_section=substr(l_image,stridx("[",l_image),strlen(l_image))
   t_image=s_image
}

status=0
# set logfile
l_logfile=logfile
if(l_logfile=="" || l_logfile==" ") {
      print("WARNING - GEMSEEING: No logfile defined. Output to screen only.")
      l_logfile=""
}

date | scan(l_struct)
printlog("----------------------------------------------------------------------------",l_logfile,l_verbose)
printlog("GEMSEEING -- "//l_struct,l_logfile,l_verbose)
printlog("",l_logfile,l_verbose)

# check access
if(!imaccess(s_image) && !imaccess(s_image//"[0]") && !imaccess(s_image//"[1]")) {
  printlog("ERROR - GEMSEEING: Image "//l_image//" not found",l_logfile,yes)
  status=1
  bye
}
# cut .imh off s_image if present 
if(substr(s_image,strlen(s_image)-3,strlen(s_image)) == ".imh") 
   s_image=substr(s_image,1,(strlen(s_image)-4))
# cut .fits off s_image if present
if(substr(s_image,strlen(s_image)-4,strlen(s_image)) == ".fits") 
   s_image=substr(s_image,1,(strlen(s_image)-5))

l_fl_update=fl_update ; l_fl_useall=fl_useall ; l_fl_keep=fl_keep
l_fl_inter=fl_inter ; l_maxnobj=maxnobj

if(l_coords=="default")
  l_coords=s_image//"_coo"
if(l_psffile=="default")
  l_psffile=s_image//"_see"

# Check of whether this is a MEF file - if it is not a fits file
# the check should not be made
# Make image name with the MEF extension name if this is a MEF file
# Use extension defined in input, otherwise use SCI extension if 
# present, otherwise extension number 1
# This can still crash in NEXTEND is not defined in header of MEF[0]
cache("imgets")
if(access(s_image//".fits")) {
  imgets(s_image//"[0]","EXTEND", >>& "/dev/null")
  if(imgets.value!="F") {
    imgets(s_image//"[0]","NEXTEND", >>& "/dev/null")
    if(imgets.value!="0") {
      l_fl_ext=yes ; l_ext=0 
# If there is an extension specified then l_section!=""
# l_section!="" may also be due to a section specified
      if(l_section!="") {
        print(l_section) | match(",",stop-) | count | scan(l_nsec)
        if(l_nsec!=0) {  # section defined, check for extension
          print(l_section) | match("\]\[",stop-) | count | scan(l_ext)
          if(l_ext!=0) {  # section and extension defined
            l_sciext=substr(l_section,stridx("[",l_section),stridx("]",l_section))
            l_section=substr(l_section,strlen(l_sciext)+1,strlen(l_section))
            print(l_section) | match(",",stop-) | count | scan(l_nsec)
            if(l_nsec==0) {
            # just switch the variables if needed, rather than exit
              l_save = l_section ; l_section = l_sciext ; l_sciext = l_save
            }
          } else {  # only section defined, set default extension
           if(imaccess(s_image//"[SCI]"))
             l_sciext="[SCI]" 
           else
             l_sciext="[1]" 
          }
        } else {
          l_ext=1  # extension defined, no section
          l_sciext=l_section
          l_section=""
        }

        l_image=s_image//".fits"//l_sciext//l_section
        t_image=s_image//".fits"//l_sciext
        s_image=s_image//l_sciext//l_section
      } else {

        if(imaccess(s_image//"[SCI]")) {
           l_image=s_image//".fits[SCI]"//l_section
           t_image=s_image//".fits[SCI]"
           s_image=s_image//"[SCI]"//l_section
        } else {
           l_image=s_image//".fits[1]"//l_section
           t_image=s_image//".fits[1]"
           s_image=s_image//"[1]"//l_section
        }
      }
    }
  }
} else
   s_image=s_image//l_section

# slightly redundant but catches all
if(!imaccess(l_image)) {
  printlog("ERROR - GEMSEEING: Image "//l_image//" not found",l_logfile,yes)
  status=1
  bye
}

l_fl_cleed=fl_cleed
l_fl_overwrite=fl_overwrite 
if(l_fl_overwrite) 
delete(l_coords//","//l_psffile//","//l_psffile//".tab,\
  "//l_psffile//"_psf.tab",verify-, >>& "dev$null")

# check if final psffile exists
if(access(l_psffile//"_psf.tab")) {
  printlog("ERROR - GEMSEEING: PSF file "//l_psffile//"_psf.tab exists",l_logfile,yes)
  status=1
  bye
}

# Get instrument (and camera), ron and gain from header if keywords are defined
l_key_inst = key_inst
if(l_key_inst!="" && l_key_inst!=" ") {
  imgets(l_image,l_key_inst, >>& "/dev/null") ; l_inst=imgets.value
  if(l_inst=="0") 
    l_inst=instrument
} else
l_inst=instrument 

if(l_inst=="NIRI") {
  l_key_camera = key_camera
  if(l_key_camera!="" && l_key_camera!=" ") {
    imgets(l_image,l_key_camera, >>& "/dev/null") ; l_camera=imgets.value
    if(l_camera=="0") 
      l_camera=camera
  } else
  l_camera=camera 

  l_inst=l_inst//"-"//l_camera
}
# disable database use if no valid instrument
l_fl_database=fl_database
if(l_inst=="" || l_inst==" ")
  l_fl_database=no

l_key_ron = key_ron ; l_key_gain=key_gain
if(l_key_ron!="" && l_key_ron!=" ") {
  imgets(l_image,l_key_ron, >>& "/dev/null") ; l_ron=real(imgets.value)
  if(l_ron==0) 
    l_ron=ron
} else
l_ron=ron 

if(l_key_gain!="" && l_key_gain!=" ") {
  imgets(l_image,l_key_gain, >>& "/dev/null") ; l_gain=real(imgets.value)
  if(l_gain==0) 
    l_gain=gain
} else
l_gain=gain 

# Use database if present
l_database=database
if(l_database=="" || l_database==" ") 
  l_fl_database=no
if(!access(l_database)) {
  l_fl_database=no
  printlog("WARNING - GEMSEEING: Cannot access database "//l_database,l_logfile,yes)
  printlog("                     Using parameter file values",l_logfile,yes)
}

l_pixscale=0
if(l_fl_database) {
  fields(database,"1-7",lines="1-",quit_if_miss-,print_file_n-) | \
  match(l_inst,stop-) | \
  scan(l_inst,l_pixscale,l_datamax,l_fwhmin,l_radius,l_buffer,l_width)
  if(l_pixscale==0) {
    printlog("WARNING - GEMSEEING: Instrument "//l_inst//" not in database",l_logfile,yes)
    printlog("          Using parameter file values",l_logfile,yes)
    goto getparam
  }
} else {
getparam:
  l_pixscale = pixscale ; l_fwhmin = fwhmin ; l_radius=radius 
  l_buffer=buffer ; l_width=width ; l_datamax = datamax
}

# Set fwhmin to pixels rather than arcsec
l_fwhmin=l_fwhmin/l_pixscale
if(l_radius==0)
  l_radius=l_fwhmin*2.5  # changed from 2 to 3*1.1
l_fwhmmin=fwhmmin/l_pixscale ; l_fwhmmax=fwhmmax/l_pixscale
if(l_fwhmmin==0 && l_fwhmmax==0) {
  l_fwhmmin=l_fwhmin/4. ; l_fwhmmax=l_fwhmin*4.
}
if(l_fwhmmax==0 || l_fwhmmax<=l_fwhmin) 
  l_fwhmmax=l_fwhmin*4.

l_fwhmnonao=fwhmnonao

# Strehl ratio - check validity in database
l_fl_strehl=fl_strehl
if(l_fl_strehl) {
  l_stdatabase=stdatabase
  if(l_stdatabase=="" || l_stdatabase==" " || !access(l_stdatabase)) {
    print("WARNING - GEMSEEING: Strehl ratio database not found, fl_strehl=no",l_logfile,yes)
    l_fl_strehl=no
  }
}

if(l_fl_strehl) {
#  l_stdatabase=stdatabase
  l_key_filter=key_filter
  if(l_key_filter!="" && l_key_filter!=" ") {
   imgets(l_image,l_key_filter, >>& "/dev/null") ; l_filter=imgets.value
   if(l_filter=="0") 
     l_filter=filter
  } else
  l_filter=filter 
  if(l_filter=="" || l_filter==" " || l_filter=="0") {
    l_fl_strehl=no
    goto skipstrehl
  }
  # attach spaces to filter, so a simple match will find it
  l_filter=" "//l_filter//" "

  l_npeak=0 ; l_wavelength=0 ; junk=""
  fields(l_stdatabase,"1-5",lines="1-",quit_if_miss-,print_file_n-) | \
  match(l_inst,stop-) | match(l_filter,stop-) | \
  scan(junk,junk,l_wavelength,l_filter,l_npeak)
  if(l_npeak==0)  {
    l_fl_strehl=no
    printlog("WARNING - GEMSEEING: Filter/instrument combination "//l_inst//" "//l_filter//" not in Strehl database",l_logfile,yes)
    printlog("                     No calculation of Strehl ratio",l_logfile,yes)
    goto skipstrehl
  }
}

skipstrehl:

l_niter=niter 
l_sharplow=sharplow ; l_sharphigh=sharphigh
l_mdaofaint=mdaofaint ; l_sharpsig=sharpsig
l_dmagsig=dmagsig ; l_fwhmsig=fwhmsig

tmpcoo = mktemp("tmpcoo")
tmpdao = mktemp("tmpdao")
tmpdat = mktemp("tmpdat")
tmpime = mktemp("tmpime")
tmpcd = mktemp("tmpcd")
tmpfit = mktemp("tmpfit")
tmpmom = mktemp("tmpmom")
tmpgra = mktemp("tmpgra")

# ---------------------------------------------------------------------
# Find the objects using daofind or interact, if l_coords does not exist already
# ---------------------------------------------------------------------

l_fl_disp=no
if(!access(l_coords) ) {

# ----- Interactive finding -----
if(l_fl_inter) {
  print("")
  print("Use the image cursor to point to objects in the display")
  print("   x  - select object")
  print("   q  - quit, when done")
  print("")
# need to display the image explicitly to make the tvmark work
  display(l_image,1, >>& "dev$null")
  imexamine(l_image,frame=1,logfile=l_coords,keeplog+,
    display="display(image'$1',frame=1)")
  l_fl_disp=yes
# ----- DAOFIND -----
} else {
  # set threshold  avoid extreme values
  imstat(l_image,fields="midpt",format=no,lower=-l_datamax,upper=l_datamax,
    binwidth=0.1) | scan(l_level)
  imstat(l_image,fields="midpt",format=no,
    lower=(l_level-5.*sqrt( (l_ron/l_gain)**2+l_level/l_gain ) ),
    upper=(l_level+5.*sqrt( (l_ron/l_gain)**2+l_level/l_gain ) ),
    binwidth=0.1 ) | scan(l_level)
  
  if(sigthres==INDEF) {
    l_threshold=threshold
    l_sigthres= l_threshold /sqrt( (l_ron/l_gain)**2+l_level/l_gain )
  } else {
    l_sigthres=sigthres
  }
  n_i=strlen(s_image)
  printf("Image  %-"//str(n_i)//"s   Sigma threshold %6.1f\n",s_image,l_sigthres) | scan(l_struct)
  printlog(l_struct,l_logfile,l_verbose)
  printf("       %-"//str(n_i)//"s   Sky level [ADU] %6.1f\n"," ",l_level) | scan(l_struct)
  printlog(l_struct,l_logfile,l_verbose)
  
  # find objects try to get rid of cosmics here
  daofind(l_image,output=l_coords,starmap="",skymap="",datapars="",
    findpars="",boundary="nearest",
    interactive=no,verify=no,update=no,
    fwhmpsf=l_fwhmin,readnoise=l_ron,epadu=l_gain,ccdread="",gain="",
    datapars.noise="poisson",exposure="",airmass="",filter="",obstime="",
    itime=1,datapars.sigma=(sqrt( (l_ron/l_gain)**2+l_level/l_gain )),
    datapars.datamax=l_datamax,datapars.datamin=-l_datamax,
    scale=1.,emission=yes,findpars.nsigma=1.5,
    findpars.threshold=l_sigthres,findpars.ratio=1.,findpars.theta=0.,
    sharplo=0.2,sharphi=1.0,roundlo=-1.,roundhi=1.,findpars.mkdetections-,
    icommands="",gcommands=""  )
} # end of finding

} else {
   printf("Image  %s\n",s_image) | scan(l_struct)
   printlog(l_struct,l_logfile,l_verbose)
   print("Coordinate file "//l_coords//" exists, daofind not run") | scan(l_struct)
   printlog(l_struct,l_logfile,l_verbose)
} # end of accessing coordinate file

# get image size in any case
imgets(l_image,"i_naxis1") ; l_Xmax = int(imgets.value)
imgets(l_image,"i_naxis2") ; l_Ymax = int(imgets.value)

# Do this check in any case
# check if coordinate file is from interactive or daofind
head(l_coords,nlines=1) | scan(junk1,junk2)
if(junk2!="IRAF" && !l_fl_inter) {
  l_fl_inter=yes
  printlog("WARNING - GEMSEEING: Existing coordinate file is not from daofind",l_logfile,yes)
}
if(junk2=="IRAF" && l_fl_inter) {
  l_fl_inter=no
  printlog("WARNING - GEMSEEING: Existing coordinate file is from daofind",l_logfile,yes)
}

if(!l_fl_inter)  # only if daofind has been run
  txdump(l_coords,"threshold",yes) | scan(l_sigthres)
else
  l_sigthres=10000.

# ---------------------------------------------------------------------
# Derive direct fwhm using psfmeasure, if l_psffile does not exist
# ---------------------------------------------------------------------
if(!access(l_psffile)) {

# ---- Interactive -----
if(l_fl_inter) {
  fields(l_coords,"1,2",lines="1-",quit-,print-, > tmpcoo)
  count(tmpcoo) | scan(l_obj)
  if(l_obj==0) {
    printlog("ERROR - GEMSEEING: No objects in coordinate file "//l_coords,l_logfile,yes)
    delete(tmpcoo//","//l_coords,verify-)
    status=1
  bye
  }

}
# ---- Non-interactive -----
else {

txdump(l_coords,"xcenter,ycenter,mag","mag<-0.05 && sharpness<999",
 headers-,parameters+, > tmpdao)

# Check number of objects
count(tmpdao) | scan(l_obj)
if(l_obj==0) {
  printlog("ERROR - GEMSEEING: No valid objects found in image "//l_image,l_logfile,yes)
  delete(tmpdao//","//l_coords,verify-)
  status=1
  bye
}

# get rid of objects too close to the edge of the image
tselect(tmpdao,tmpcoo,
  "c1>5 && c1<("//l_Xmax//"-5) && c2>5 && c2<("//l_Ymax//"-5)")
delete(tmpdao,verify=no)

# Count objects  
count(tmpcoo) | scan(l_obj)
if(l_obj==0) {
  printlog("ERROR - GEMSEEING: No valid objects found in image "//l_image,l_logfile,yes)
  delete(tmpdao//","//tmpcoo,verify-)
  status=1
  bye
}

n_i=strlen(s_image)
printf("Image  %-"//str(n_i)//"s   Total Nobj      %6d\n",s_image,l_obj) | scan(l_struct)
printlog(l_struct,l_logfile,l_verbose)

# If fl_useall==no:  Select subset 
if(!l_fl_useall) {
# If there are more than l_maxnobj objects cut 0.5 mag fainter than
# mdaofaint unless that is larger than zero
  if(l_obj>l_maxnobj && l_mdaofaint<-0.5) {
   tselect(tmpcoo,tmpdao,"c3 < "//str(l_mdaofaint+0.5))
   delete(tmpcoo,verify=no)
   rename(tmpdao,tmpcoo,field="all")
  }
# If there are still more than l_maxnobj objects select subset
  count(tmpcoo) | scan(l_obj)
  if(l_obj>l_maxnobj ) {
    l_obj=max(2,int(l_obj/real(l_maxnobj)+0.5))
    tselect(tmpcoo,tmpdao,"mod(row(),"//str(l_obj)//")==0")
    delete(tmpcoo,verify=no)
    rename(tmpdao,tmpcoo,field="all")
  }

} 

} # end of non-interactive

count(tmpcoo) | scan(l_obj)
n_i=strlen(s_image)
printf("       %-"//str(n_i)//"s   Fitting Nobj    %6d\n"," ",l_obj) | scan(l_struct)
printlog(l_struct,l_logfile,l_verbose)
printf("       %-"//str(n_i)//"s   Radius [pixels] %6.2f\n"," ",l_radius) | scan(l_struct)
printlog(l_struct,l_logfile,l_verbose)

printf(":show "//l_psffile//"\n:size MFWHM\n:show "//l_psffile//"\nq\n", > tmpgra)

# if l_obj=1 need to make fake line in input file
# this in fact works ...
if(l_obj==1) {
  type(tmpcoo, >> tmpcoo)
}

# Use pixelscale 1, convert later ...
# psfmeasure silently throws away any objects it cannot handle, without 
# putting an entry in the output file ....
#
# For debugging :
# psfmeasure(l_image,coords="markall",wcs="logical",display-,frame=1,
#   level=0.5,size="FWHM",scale=1.,radius=l_radius,sbuffer=l_buffer,
#   swidth=l_width,saturation=l_datamax,ignore_sat+,iterations=l_niter,
#   logfile=l_psffile,imagecur=tmpcoo,graphcur="")
# bye
# psfmeasure crashes if ignore_sat+ and only saturated input objects
psfmeasure(l_image,coords="markall",wcs="logical",display-,frame=1,
  level=0.5,size="FWHM",beta=INDEF,scale=1.,radius=l_radius,sbuffer=l_buffer,
  swidth=l_width,saturation=l_datamax,ignore_sat-,iterations=l_niter,
  logfile="",xcenter=INDEF,ycenter=INDEF,
  imagecur=tmpcoo,graphcur=tmpgra, >>& "/dev/null", >>G "/dev/null")
delete(tmpgra//","//tmpcoo,verify-)
} else {
    printlog("PSF file "//l_psffile//" exists, fitting not run",l_logfile,l_verbose)
}

# ---------------------------------------------------------------------
# Make psffile//".tab" if it does not already exist 
# ---------------------------------------------------------------------

if(!access(l_psffile//".tab")) {

if(l_fl_inter) {
   printf("x_dao r f6.2\ny_dao r f6.2\n") | \
   tcreate(tmpdao,"STDIN",l_coords,upar="",nskip=0,nlines=0,nrows=0,
    hist-,extrapar=5,tbltype="default",extracol=0)
} else {
# make STSDAS table with output from psfmeasure and daofind
   pconvert(l_coords,tmpdao,"*",expr="yes",append-)
   tchcol(tmpdao,"xcenter","x_dao","","",verbose-)
   tchcol(tmpdao,"ycenter","y_dao","","",verbose-)
   tchcol(tmpdao,"mag","m_dao","","",verbose-)
}

# make the psffile STSDAS table from psfmeasure output

printf("mfwhm r f7.3\nbeta r f6.2\n", > tmpcd)
printf("x_psf r f7.2\ny_psf r f7.2\nm_psf r f7.3\nfwhm r f7.3\neps r f7.3\npos r f7.3\n", >> tmpcd)
print("sat ch*1 %s", >> tmpcd)

# Find the length of psffile; has nothing to do with number of input objects
count(l_psffile) | scan(l_obj) ; l_obj=l_obj/2

if(l_obj==0) {
  printlog("ERROR - GEMSEEING: No valid objects fitted in image "//l_image,l_logfile,yes)
  delete(tmpdao//".tab,"//tmpcd,verify-)
  status=1
  bye
}

# For each get rows:  l_obj/2 excl. first 3 and last 2
# There are saturated objects in this because psfmeasure crashes
# if only saturated input objects and run with ignore_sat+
for(n_i=1;n_i<=2;n_i+=1) {
  if(n_i==1) {
  fields(l_psffile,"2-8",lines=str((n_i-1)*l_obj+4), > tmpdat)
  fields(l_psffile,"1-7",
   lines=str((n_i-1)*l_obj+5)//"-"//str((n_i*l_obj-2)), >> tmpdat)
  } else {
     fields(l_psffile,"5,6",lines=str((n_i-1)*l_obj+4), > tmpgra)
     fields(l_psffile,"4,5",
     lines=str((n_i-1)*l_obj+5)//"-"//str((n_i*l_obj-2)), >> tmpgra)
     join(tmpgra,tmpdat,output=tmpmom,delim=" ",missing="Missing",maxchars=161,
      shortest+,verbose+)
     delete(tmpdat//","//tmpgra,verify-)
     rename(tmpmom,tmpdat,field="all")
  }
}

# need to change how saturated objects are marked since tselect cannot
# handle this silly output from psfmeasure

sed("-e 's/$/N/'",tmpdat," | sed -e 's/\*   N/Y/' ", >> tmpgra)

tcreate(tmpime,tmpcd,tmpgra,upar="",nskip=0,nlines=1,nrows=0,hist-,
 extrapar=5,tbltype="default",extracol=0)
delete(tmpgra//","//tmpdat//","//tmpcd,verify=no)

# Get rid of objects that are not unique. There such objects because
# daofind finds lots of spurious objects in AO images.
# This does not take care of all of them, but a large part of them
tproject(tmpime,tmpgra,"*",uniq+)

# match the two tables
if(l_fl_inter) {
tmatch(tmpgra,tmpdao,l_psffile//".tab","x_psf,y_psf","x_dao,y_dao",
 maxnorm=(l_fwhmin*3.0),incol1="",incol2="",factor="",diagfile="",nmcol1="",
 nmcol2="",sphere-)

} else {
tmatch(tmpgra,tmpdao,l_psffile//".tab","x_psf,y_psf","x_dao,y_dao",
 maxnorm=(l_fwhmin*0.5),incol1="",incol2="",factor="",diagfile="",nmcol1="",
 nmcol2="",sphere-)
 # calculate magnitude difference
 tcalc(l_psffile//".tab","dmag","m_dao-m_psf",colfmt="f6.3",colunits="")
}
tdelete(tmpdao//","//tmpime//","//tmpgra,verify=no)

} else {
    printlog("PSF STSDAS table "//l_psffile//".tab exists.",l_logfile,l_verbose)
}

# -----------------------------------------------------------------
# Select objecs from the table to be used for seeing determination
# If there after the first selection is less than 5 objects left
# the program goes directly to the final run of psfmeasure and the 
# statistics. If there after any of the following selections is less
# than 3 objects left, that selection is omitted and the final run
# of psfmeasure and statistics is done on the sample before that last
# selection.
# -----------------------------------------------------------------
cache("tselect","tstat","tinfo")

# check if l_psffile.tab has a dmag column 
if(!l_fl_inter) {
  l_col=0
  tlcol(l_psffile//".tab",nlist=1) | match("dmag","STDIN") | count | scan(l_col)
  if(l_col==0) {
    print("WARNING: Existing PSF file is from interactive use.")
    print("         Selection turned off")
    l_fl_inter=yes
  }
}

# use all if interactive or if dmag column was missing
if(l_fl_inter) {
  tcalc(l_psffile//".tab","flag","1",colfmt="i2",colunit="")
} else {

# select on saturated and magnitude
tselect(l_psffile//".tab",tmpdao,
  "sat!='Y' && m_dao<"//l_mdaofaint//" && fwhm>="//l_fwhmmin//" && fwhm<"//l_fwhmmax )
tinfo(tmpdao,ttout=no)
if(tinfo.nrows<3) {
  printlog("WARNING - GEMSEEING: Too few objects, only saturated objects and objects",l_logfile,yes)
  printf("                     with fwhm < %7.4f pixels are excluded\n",l_fwhmmin) | scan(l_struct)
  printlog(l_struct,l_logfile,yes)
  tdelete(tmpdao,verify=no)
  tselect(l_psffile//".tab",tmpdao,"sat!='Y' && fwhm>="//l_fwhmmin)
  tinfo(tmpdao,ttout-)
  if(tinfo.nrows<1) {
    printlog("ERROR - GEMSEEING: All objects saturated",l_logfile,yes)
    tdelete(tmpdao,verify-)
    status=1
    bye
  }
  goto finalselection
}
if(tinfo.nrows<=5)
  goto finalselection

tstat(tmpdao,"dmag",outtable="",lowlim=INDEF,highlim=INDEF,rows="-") 
tselect(tmpdao,tmpime,"m_psf<("//l_mdaofaint//"-"//tstat.median//")" )
tdelete(tmpdao,verify=no)

tinfo(tmpime,ttout=no)
if(tinfo.nrows<3) {
  tdelete(tmpime,verify=no)
  goto finalselection
}

# select on sigma deviation in sharp
tstat(tmpime,"sharpness",outtable="",lowlim=l_sharplow,highlim=l_sharphigh,
  rows="-") 
if(tstat.stddev>0) {
 l_expression="sharpness>=("//tstat.median//"-"//l_sharpsig//"*"//tstat.stddev//") && sharpness<=("//tstat.median//"+"//l_sharpsig//"*"//tstat.stddev//")"
 tselect(tmpime,tmpdao,l_expression)
 tdelete(tmpime,verify=no)
} else {
 printlog("WARNING - GEMSEEING: No selection on sharpness",l_logfile,yes)
 rename(tmpime//".tab",tmpdao//".tab",field="all")
}

tinfo(tmpdao,ttout=no)
if(tinfo.nrows<3) {
  tdelete(tmpdao,verify=no)
  rename(tmpime//".tab",tmpdao//".tab",field="all")
  goto finalselection
}

# select on sigma deviation in dmag
tstat(tmpdao,"dmag",outtable="",lowlim=INDEF,highlim=INDEF,rows="-")
if(tstat.stddev>0) {
  l_expression="dmag>=("//tstat.median//"-"//l_dmagsig//"*"//tstat.stddev//") && dmag<=("//tstat.median//"+"//l_dmagsig//"*"//tstat.stddev//")"
  tselect(tmpdao,tmpime,l_expression) ; tdelete(tmpdao,verify=no)
} else {
 printlog("WARNING - GEMSEEING: No selection on rms dmag",l_logfile,yes)
 rename(tmpdao//".tab",tmpime//".tab",field="all")
}

tinfo(tmpime,ttout=no)
if(tinfo.nrows<3) {
  tdelete(tmpime,verify=no)
  goto finalselection
}

# select on sigma deviation in fwhm
tstat(tmpime,"fwhm",outtable="",lowlim=INDEF,highlim=INDEF,rows="-")
if(tstat.stddev>0) {
  l_expression="fwhm>=("//tstat.median//"-"//l_fwhmsig//"*"//tstat.stddev//") && fwhm<=("//tstat.median//"+"//l_fwhmsig//"*"//tstat.stddev//")"
  tselect(tmpime,tmpdao,l_expression) ; tdelete(tmpime,verify=no)
} else {
 printlog("WARNING - GEMSEEING: No selection on rms fwhm",l_logfile,yes)
 rename(tmpime//".tab",tmpdao//".tab",field="all")
}

tinfo(tmpdao,ttout=no)
if(tinfo.nrows<3) {
  tdelete(tmpdao,verify=no)
  rename(tmpime//".tab",tmpdao//".tab",field="all")
  goto finalselection
}

finalselection:
# Resulting direct FWHM 
tstat(tmpdao,"fwhm",outtable="",lowlim=INDEF,highlim=INDEF,rows="-")
n_i=strlen(s_image)
printf("Image  %-"//str(n_i)//"s   FWHM [arcsec]  %7.4f\n",s_image,tstat.median*l_pixscale) | scan(l_struct)
printlog(l_struct,l_logfile,l_verbose)

# mark objects in l_psffile//".tab" that have been used

# objects not used
tdiff(l_psffile//".tab",tmpdao,tmpime,"id","id")
tinfo(tmpime,ttout=no)
if(tinfo.nrows!=0)
  tcalc(tmpime,"flag","0",colfmt="i2",colunit="")
# objects used
tcalc(tmpdao,"flag","1",colfmt="i2",colunit="")
tdelete(l_psffile//".tab",verify=no)

tmerge(tmpdao//","//tmpime,l_psffile//".tab","append",allcols+,
  tbltype="default")
tsort(l_psffile//".tab","id",ascend+,casesens+)

tdelete(tmpdao//","//tmpime,verify=no)
delete(tmpcoo,verify=no, >>& "/dev/null")

}  # end of selection

# ----------------------------------------------------------------------
# Rerun psfmeasure on much fewer objects to get 
#  FWHM, EED50, EED85 and MFWHM
# ----------------------------------------------------------------------

tselect(l_psffile//".tab",tmpdao,"flag==1")
tprint(tmpdao,prparam-,prdata+,pwidth=80,
 col="x_psf,y_psf",showr-,showh-,orig_row+,showu-,rows="-",
 option="plain",align+,sp_col="",lgroup=0, > tmpcoo)
#print("q", >> tmpcoo)

# count objects and make two lines if needed
l_fl_one=no
count(tmpcoo) | scan(l_obj)
if(l_obj==1) {
  type(tmpcoo, >> tmpcoo)
  l_fl_one=yes
}

# Change level before size or level will not get changed ...
printf(":level 0.5\n:size Radius\n:show "//tmpime//"\n", >> tmpgra)
printf(":level 0.85\n:size Radius\n:show "//tmpime//"\nq\n", >> tmpgra)

# iterations=1 since radius should be fixed in this step if it is AO observations
# The check on AO or not should be changed later to look for some Altair keyword
if(l_inst=="NIRI-F/32" || l_inst=="Hokupaa/QUIRC" || l_inst=="Hokupaa/QUIRC/Ded" \
   || l_inst=="AO-SIM") {
psfmeasure(l_image,coords="markall",wcs="logical",display-,frame=1,
  level=0.5,size="FWHM",beta=INDEF,scale=1.,radius=(l_fwhmnonao*3./l_pixscale),
  sbuffer=(l_fwhmnonao*2./l_pixscale),
  swidth=l_width,saturation=l_datamax,ignore_sat+,iterations=1,
  logfile="",xcenter=INDEF,ycenter=INDEF,
  imagecur=tmpcoo,graphcur=tmpgra, >> "/dev/null", >>G "/dev/null")

} else {
psfmeasure(l_image,coords="markall",wcs="logical",display-,frame=1,
  level=0.5,size="FWHM",beta=INDEF,scale=1.,radius=l_radius,sbuffer=l_buffer,
  swidth=l_width,saturation=l_datamax,ignore_sat+,iterations=l_niter,
  logfile="",xcenter=INDEF,ycenter=INDEF,
  imagecur=tmpcoo,graphcur=tmpgra, >> "/dev/null", >>G "/dev/null")
}
delete(tmpgra//","//tmpcoo,verify-)

# make the psffile STSDAS table from psfmeasure output
# need to get rid of the image name in the file ...
printf("x_psf r f7.2\ny_psf r f7.2\n", > tmpcd)
printf("eed50 r f7.3\need85 r f7.3\n", >> tmpcd)

# Find the length of this psffile
count(tmpime) | scan(l_obj) ; l_obj=l_obj/2

# For each get rows:  l_obj/2 excl. first 3 and last 2
# There should be no saturated objects in this
for(n_i=1;n_i<=2;n_i+=1) {
  if(n_i==1) {
  fields(tmpime,"2-3,5",lines=str((n_i-1)*l_obj+4), > tmpdat)
  fields(tmpime,"1-2,4",
   lines=str((n_i-1)*l_obj+5)//"-"//str((n_i*l_obj-2)), >> tmpdat)
  } else {
     fields(tmpime,"5",lines=str((n_i-1)*l_obj+4), > tmpgra)
     fields(tmpime,"4",
     lines=str((n_i-1)*l_obj+5)//"-"//str((n_i*l_obj-2)), >> tmpgra)
     join(tmpdat,tmpgra,output=tmpmom,delim=" ",missing="Missing",maxchars=161,
      shortest+,verbose+)
     delete(tmpdat//","//tmpgra,verify-)
     rename(tmpmom,tmpdat,field="all")
  }
}
delete(tmpime,verify-)
if(l_fl_one) {
  tcreate(tmpcoo,tmpcd,tmpdat,upar="",nskip=0,nlines=1,
    nrows=0,hist-,extrapar=5,tbltype="default",extracol=0)
  tselect(tmpcoo,tmpime,"row()==1")
  tdelete(tmpcoo,verify-)
} else {
  tcreate(tmpime,tmpcd,tmpdat,upar="",nskip=0,nlines=1,
    nrows=0,hist-,extrapar=5,tbltype="default",extracol=0)
}
tmerge(tmpime//","//tmpdao,l_psffile//"_psf.tab","merge",allcols+,
  tbltype="default")
delete(tmpcd//","//tmpdat//","//tmpdao//".tab,"//tmpime//".tab",verify-)

# convert all measurements from pixels to arcsec, eed?? to diameters
tcalc(l_psffile//"_psf.tab","fwhm","fwhm*"//str(l_pixscale),
  colfmt="f7.4",colunit="")
tcalc(l_psffile//"_psf.tab","mfwhm","mfwhm*"//str(l_pixscale),
  colfmt="f7.4",colunit="")
tcalc(l_psffile//"_psf.tab","eed50","eed50*2*"//str(l_pixscale),
  colfmt="f7.4",colunit="")
tcalc(l_psffile//"_psf.tab","eed85","eed85*2*"//str(l_pixscale),
  colfmt="f7.4",colunit="")

# Final determinations - median values
if(l_fl_keep || l_fl_update)
  date | scan(l_struct)
# put keywords in STSDAS table

if(l_fl_keep) {
parkey(l_struct,l_psffile//"_psf.tab","GEMSEE",add=yes)
if(l_fl_strehl)
  parkey("gemseeing:  psfmeasure  direct/eed50/eed85/moffat/strehl",
    l_psffile//"_psf.tab","SEEPRG", add=yes)
else
  parkey("gemseeing:  psfmeasure  direct/eed50/eed85/moffat",
    l_psffile//"_psf.tab","SEEPRG", add=yes)

parkey(l_image,l_psffile//"_psf.tab","IMAGE",add=yes)
parkey(l_image,l_psffile//"_psf.tab","SECTION",add=yes)
parkey(l_inst,l_psffile//"_psf.tab","INSTRUME",add=yes)
parkey(l_fwhmin,l_psffile//"_psf.tab","FWHM_IN",add=yes)
parkey(l_pixscale,l_psffile//"_psf.tab","PIXSCALE",add=yes)
parkey(l_ron,l_psffile//"_psf.tab","RON",add=yes)
parkey(l_gain,l_psffile//"_psf.tab","EPADU",add=yes)
parkey(l_sigthres,l_psffile//"_psf.tab","SIGTHRES",add=yes)
parkey("psfmeasure",l_psffile//"_psf.tab","FITPRG",add=yes)
parkey("direct",l_psffile//"_psf.tab","FITTYP",add=yes)
parkey(l_radius,l_psffile//"_psf.tab","FITRAD",add=yes)
parkey(l_buffer,l_psffile//"_psf.tab","FITBUF",add=yes)
parkey(l_width,l_psffile//"_psf.tab","FITWID",add=yes)
parkey(l_mdaofaint,l_psffile//"_psf.tab","MDAOFAI",add=yes)
parkey(l_datamax,l_psffile//"_psf.tab","DATAMAX",add=yes)
parkey(l_sharplow,l_psffile//"_psf.tab","SHARPLO",add=yes)
parkey(l_sharphigh,l_psffile//"_psf.tab","SHARPHI",add=yes)
parkey(l_sharpsig,l_psffile//"_psf.tab","SHARPSIG",add=yes)
parkey(l_dmagsig,l_psffile//"_psf.tab","DMAGSIG",add=yes)
parkey(l_fwhmsig,l_psffile//"_psf.tab","FWHMSIG",add=yes)
}

tstat(l_psffile//"_psf.tab","fwhm",outtable="",lowlim=INDEF,highlim=INDEF)
out_fwhm=tstat.median
out_npsf=tstat.nrows
if(out_npsf>2)
 rms_fwhm=tstat.stddev
else
 rms_fwhm=0.

if(l_fl_keep) {
parkey(tstat.median,l_psffile//"_psf.tab","FWHMPSF",add=yes)
parkey(tstat.mean,l_psffile//"_psf.tab","FWHMMEAN",add=yes)
parkey(tstat.median,l_psffile//"_psf.tab","FWHMMED",add=yes)
parkey(tstat.stddev,l_psffile//"_psf.tab","FWHMRMS",add=yes)
parkey(tstat.vmin,l_psffile//"_psf.tab","FWHMMIN",add=yes)
parkey(tstat.vmax,l_psffile//"_psf.tab","FWHMMAX",add=yes)
parkey(tstat.nrows,l_psffile//"_psf.tab","NPSF",add=yes)
}
if(l_fl_update) {
  gemhedit(l_image,"GEMSEE",l_struct,"Time stamp for gemseeing")
  if(l_fl_strehl)
    hedit(l_image,"SEEPRG",
      "gemseeing:  psfmeasure  direct/eed50/eed85/moffat/strehl",
      add=yes,verify=no,update=yes,show-,delete=no)
  else
    hedit(l_image,"SEEPRG","gemseeing:  psfmeasure  direct/eed50/eed85/moffat",
      add=yes,verify=no,update=yes,show-,delete=no)
  gemhedit(l_image,"FWHMPSF",tstat.median,"FWHM of PSF [arcsec], direct measurement")
  gemhedit(l_image,"NPSF",tstat.nrows,
    "Number of stars used by gemseeing")
}
tstat(l_psffile//"_psf.tab","mfwhm",outtable="",lowlim=INDEF,highlim=INDEF)
out_mfwhm=tstat.median
rms_mfwhm=tstat.stddev
if(l_fl_keep) {
parkey(tstat.median,l_psffile//"_psf.tab","MFWHM",add=yes)
parkey(tstat.mean,l_psffile//"_psf.tab","MFWHMMEA",add=yes)
parkey(tstat.median,l_psffile//"_psf.tab","MFWHMMED",add=yes)
parkey(tstat.stddev,l_psffile//"_psf.tab","MFWHMRMS",add=yes)
parkey(tstat.vmin,l_psffile//"_psf.tab","MFWHMMIN",add=yes)
parkey(tstat.vmax,l_psffile//"_psf.tab","MFWHMMAX",add=yes)
}
if(l_fl_update)
  gemhedit(l_image,"MFWHM",tstat.median,"FWHM of PSF [arcsec], Moffat fit")

tstat(l_psffile//"_psf.tab","eed50",outtable="",lowlim=INDEF,highlim=INDEF)
# clean the measurement if enough objects
if(tstat.nrows>15 && l_fl_cleed) {
  tstat(l_psffile//"_psf.tab","eed50",outtable="",lowlim=INDEF,highlim=(1.5*tstat.median))
  tstat(l_psffile//"_psf.tab","eed50",outtable="",lowlim=INDEF,highlim=(1.5*tstat.median))
}
out_eed50=tstat.median
rms_eed50=tstat.stddev
if(l_fl_keep) {
parkey(tstat.median,l_psffile//"_psf.tab","EED50",add=yes)
parkey(tstat.mean,l_psffile//"_psf.tab","EED50MEA",add=yes)
parkey(tstat.median,l_psffile//"_psf.tab","EED50MED",add=yes)
parkey(tstat.stddev,l_psffile//"_psf.tab","EED50RMS",add=yes)
parkey(tstat.vmin,l_psffile//"_psf.tab","EED50MIN",add=yes)
parkey(tstat.vmax,l_psffile//"_psf.tab","EED50MAX",add=yes)
}
if(l_fl_update)
  gemhedit(l_image,"EED50",tstat.median,"50-percent encircled energy diameter [arcsec]")

tstat(l_psffile//"_psf.tab","eed85",outtable="",lowlim=INDEF,highlim=INDEF)
# clean the measurement if enough objects
if(tstat.nrows>15 && l_fl_cleed) {
  tstat(l_psffile//"_psf.tab","eed85",outtable="",lowlim=INDEF,highlim=(1.5*tstat.median))
  tstat(l_psffile//"_psf.tab","eed85",outtable="",lowlim=INDEF,highlim=(1.5*tstat.median))
}
out_eed85=tstat.median
rms_eed85=tstat.stddev
if(l_fl_keep) {
parkey(tstat.median,l_psffile//"_psf.tab","EED85",add=yes)
parkey(tstat.mean,l_psffile//"_psf.tab","EED85MEA",add=yes)
parkey(tstat.median,l_psffile//"_psf.tab","EED85MED",add=yes)
parkey(tstat.stddev,l_psffile//"_psf.tab","EED85RMS",add=yes)
parkey(tstat.vmin,l_psffile//"_psf.tab","EED85MIN",add=yes)
parkey(tstat.vmax,l_psffile//"_psf.tab","EED85MAX",add=yes)
}
if(l_fl_update)
  gemhedit(l_image,"EED85",tstat.median,"85-percent encircled energy diameter [arcsec]")

tstat(l_psffile//"_psf.tab","eps",outtable="",lowlim=INDEF,highlim=INDEF)
out_eps=tstat.median
rms_eps=tstat.stddev
if(l_fl_keep) {
parkey(tstat.median,l_psffile//"_psf.tab","EPSPSF",add=yes)
parkey(tstat.mean,l_psffile//"_psf.tab","EPSMEAN",add=yes)
parkey(tstat.median,l_psffile//"_psf.tab","EPSMED",add=yes)
parkey(tstat.stddev,l_psffile//"_psf.tab","EPSRMS",add=yes)
parkey(tstat.vmin,l_psffile//"_psf.tab","EPSMIN",add=yes)
parkey(tstat.vmax,l_psffile//"_psf.tab","EPSMAX",add=yes)
}
if(l_fl_update)
  gemhedit(l_image,"EPSPSF",tstat.median,"Ellipticity of PSF")

tstat(l_psffile//"_psf.tab","pos",outtable="",lowlim=INDEF,highlim=INDEF)
out_pos=tstat.median
rms_pos=tstat.stddev
if(l_fl_keep) {
parkey(tstat.median,l_psffile//"_psf.tab","POSPSF",add=yes)
parkey(tstat.mean,l_psffile//"_psf.tab","POSMEAN",add=yes)
parkey(tstat.median,l_psffile//"_psf.tab","POSMED",add=yes)
parkey(tstat.stddev,l_psffile//"_psf.tab","POSRMS",add=yes)
parkey(tstat.vmin,l_psffile//"_psf.tab","POSMIN",add=yes)
parkey(tstat.vmax,l_psffile//"_psf.tab","POSMAX",add=yes)
}
if(l_fl_update)
  gemhedit(l_image,"POSPSF",tstat.median,"Position angle of PSF, CCW from X-axis")

# --------------------------------------------------------------------
# Determination of approximative Strehl ratio if fl_strehl=yes
# --------------------------------------------------------------------

if(l_fl_strehl) {
  tprint(l_psffile//"_psf.tab",col="x_psf,y_psf",showr-,showh-,prdata+,
   prparam-,pwidth=80,plength=0,orig_row+,showunits-,rows="-",
   option="plain",align+,sp_col="",lgroup=0, > tmpcoo)
  phot(l_image,skyfile="",coords=tmpcoo,output=tmpdao,plotfile="",
   datapars="",centerpars="",fitskypars="",photpars="",interactive-,
   radplots-,verify-,update-,verbose-,icommands="",gcommands="",
   datapars.scale=1,datapars.emission+,datapars.datamin=-l_datamax,
   datapars.datamax=l_datamax,noise="poisson",datapars.ccdread="",
   datapars.gain="",datapars.readnoise=l_ron,datapars.epadu=l_gain,
   datapars.exposure="",datapars.airmass="",datapars.filter="",
   datapars.obstime="",datapars.itime=1,centerpars.calgorithm="none",
   fitskypars.salgorithm="gauss",
   fitskypars.annulus=(out_eed85/l_pixscale+l_buffer),
   fitskypars.dannulus=l_width,fitskypars.smaxiter=10,
   fitskypars.sloclip=0,fitskypars.shiclip=0,fitskypars.snreject=50,
   fitskypars.sloreject=3,fitskypars.shireject=3,fitskypars.khist=5,
   fitskypars.binsize=0.1,fitskypars.smooth-,fitskypars.rgrow=0,
   fitskypars.mksky-,photpars.weighting="constant",
   photpars.apertures=str(out_eed50/l_pixscale/2.),photpars.zmag=25.,
   photpars.mkapert-)

   pconvert(tmpdao,tmpime,"flux,mag,merr,msky",expr="yes",append-)
   delete(tmpdao//","//tmpcoo,verify-)
   tmerge(l_psffile//"_psf.tab,"//tmpime,tmpdao,"merge")
   tdelete(l_psffile//"_psf.tab,"//tmpime,verify-)
   rename(tmpdao//".tab",l_psffile//"_psf.tab",field="all")

# The approximate Strehl ratio
# use the value in the central pixel plus a statistical correction
   tinfo(l_psffile//"_psf.tab",ttout-) ; l_obj=tinfo.nrows
   for(ii=1;ii<=l_obj;ii+=1) {
     tprint(l_psffile//"_psf.tab",col="x_psf,y_psf",showr-,showh-,
      showu-,option="plain",row=str(ii),prpar-,prdat+) | \
       scan(x_psf,y_psf)
     listpix(t_image//"["//int(x_psf+0.5)//","//int(y_psf+0.5)//"]") | \
        fields("STDIN","2",lines="1",quit-,print-, >> tmpdao)
   }
   printf("peak r f8.1\n") | tcreate(tmpdao//".tab","STDIN",tmpdao,
     upar="",nskip=0,nlines=0,nrows=0,hist-,extrapar=5,
     tbltype="default",extracol=0)
   tmerge(l_psffile//"_psf.tab,"//tmpdao//".tab",tmpcoo//".tab",
    "merge",allcols+,tbltype="default",extracol=0)
   delete(l_psffile//"_psf.tab,"//tmpdao//".tab,"//tmpdao,ver-)
   rename(tmpcoo//".tab",l_psffile//"_psf.tab",field="all")
# Statistical correction of 1.08 for objects not being centered on 
# pixels - approx correct for Hokupaa/QUIRC, need database info later
   tcalc(l_psffile//"_psf.tab","npeak","1.08*(peak-msky)/(flux*2.)",colfmt="f7.4")
   tcalc(l_psffile//"_psf.tab","strehl","npeak/"//str(l_npeak),colfmt="f7.4")

   tstat(l_psffile//"_psf.tab","strehl",outtable="",lowlim=INDEF,highlim=INDEF)
   out_strehl=tstat.median
   rms_strehl=tstat.stddev
 if(l_fl_keep) {
   parkey(tstat.median,l_psffile//"_psf.tab","STREHLRA",add=yes)
   parkey(tstat.mean,l_psffile//"_psf.tab","STRMEAN",add=yes)
   parkey(tstat.median,l_psffile//"_psf.tab","STRMED",add=yes)
   parkey(tstat.stddev,l_psffile//"_psf.tab","STRRMS",add=yes)
   parkey(tstat.vmin,l_psffile//"_psf.tab","STRMIN",add=yes)
   parkey(tstat.vmax,l_psffile//"_psf.tab","STRMAX",add=yes)
 }
 if(l_fl_update)
   gemhedit(l_image,"STREHLRA",tstat.median,"Strehl ratio")
 
}

# Output result
printlog("",l_logfile,yes)
n_i=strlen(s_image) ; n_i=max(n_i,8)
if(l_fl_strehl) {
  printf("%-"//str(n_i)//"s  %4d %7s %7s %7s %7s %7s %6s %5s\n",
    "Image","NPSF","FWHM","MFWHM","EED50%","EED85%","STREHL","EPS","POS") | scan(l_struct)
  printlog(l_struct,l_logfile,yes)
  printf("%-"//str(n_i)//"s  %4d %7.4f %7.4f %7.4f %7.4f %7.4f %6.3f %5.1f\n",
    s_image,out_npsf,out_fwhm,out_mfwhm,out_eed50,out_eed85,out_strehl,
    out_eps,out_pos) | scan(l_struct)
  printlog(l_struct,l_logfile,yes)
  printf("%-"//str(n_i)//"s  %4d %7.4f %7.4f %7.4f %7.4f %7.4f %6.3f %5.1f\n",
    "[stddev]","",rms_fwhm,rms_mfwhm,rms_eed50,rms_eed85,rms_strehl,
    rms_eps,rms_pos) | scan(l_struct)
  printlog(l_struct,l_logfile,yes)
} else {
  printf("%-"//str(n_i)//"s  %4d %7s %7s %7s %7s %6s %5s\n",
    "Image","NPSF","FWHM","MFWHM","EED50%","EED85%","EPS","POS") | scan(l_struct)
  printlog(l_struct,l_logfile,yes)
  printf("%-"//str(n_i)//"s  %4d %7.4f %7.4f %7.4f %7.4f %6.3f %5.1f\n",
    s_image,out_npsf,out_fwhm,out_mfwhm,out_eed50,out_eed85,out_eps,out_pos) | scan(l_struct)
  printlog(l_struct,l_logfile,yes)
  printf("%-"//str(n_i)//"s  %4d %7.4f %7.4f %7.4f %7.4f %6.3f %5.1f\n",
    "[stddev]","",rms_fwhm,rms_mfwhm,rms_eed50,rms_eed85,rms_eps,rms_pos) | scan(l_struct)
  printlog(l_struct,l_logfile,yes)
}

printlog("",l_logfile,yes)
printf("%-"//str(n_i)//"s  %s %7.4f %s\n"," ","Pixel scale = ",l_pixscale,"arcsec/pixel") | scan(l_struct)
printlog(l_struct,l_logfile,yes)

if(l_fl_inter) {
 print("")
 tprint(l_psffile//"_psf.tab",
  col="x_psf,y_psf,fwhm,mfwhm,eed50,eed85,strehl,eps,pos",showr-,showh+,
  showunits-,rows="-",option="plain",align+,sp_col="",lgroup=0)
}

if(fl_inter) {
 if(!l_fl_disp)
  display(l_image,1, >>& "dev$null")
 tprint(l_psffile//"_psf.tab",col="x_psf,y_psf,fwhm",showr-,showh-,
  showunits-,rows="-",option="plain",align+,sp_col="",lgroup=0) | \
  tvmark(1,"STDIN",autolog-,mark="point",color=204,label=yes,nxoff=5,
  nyoff=5,txsize=1,interactive=no)
}

if(!l_fl_keep)
  delete(l_coords//","//l_psffile//","//l_psffile//".tab,"//l_psffile//"_psf.tab",verify-)

end
